"""
Export functionality, wrapped by CLI and can be used programmatically.
"""

import os
import importlib
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Iterable, cast

import logging

from glyphsynth import BaseGlyph

__all__ = [
    "RASTER_SUPPORT",
    "ExportSpec",
    "export_glyphs",
]

type GlyphObjType = type[BaseGlyph] | BaseGlyph | ExportSpec
"""
Represents a glyph to be exported. If a BaseGlyph subclass or instance is
provided, an ExportSpec is automatically created using the output path.
"""

type GlyphSpecType = GlyphObjType | Iterable[GlyphSpecType] | Callable[
    [], GlyphSpecType
]
"""
Recursive type to represent an object which can be handled by exporter.
"""

INDENT = 4


@dataclass
class ExportSpec:
    """
    Container for a glyph instance and path in which to place the exported
    artifact.
    """

    glyph: BaseGlyph
    path: Path
    module: str | None = None


def export_glyphs(
    fqcn: str,
    output_path: Path,
    output_modpath: bool = False,
    svg: bool = False,
    png: bool = False,
    in_place_raster: bool = False,
):
    """
    Export all glyphs from the object imported from the fully-qualified
    class name, which may be any of the following:

    - ExportSpec
    - BaseGlyph subclass or instance
    - Iterable
    - Callable
    - Module with symbol names provided via `__all__`

    The object imported from the FQCN is recursed to collect all glyph objects.
    If a `BaseGlyph` is encountered,

    """

    logging.info(f"Exporting to output path: {output_path}")

    containers: list[ExportSpec] = _extract_containers(fqcn)

    for container in containers:
        export_path: Path

        if output_modpath:
            # if enabled, include glyph's modpath in output path hierarchy
            module = container.module or container.glyph.__module__
            export_path = (
                output_path / module.replace(".", "/") / container.path
            )
        else:
            export_path = output_path / container.path

        _export_glyph(
            container.glyph,
            export_path,
            svg,
            png,
            in_place_raster,
        )


def _export_glyph(
    glyph: BaseGlyph,
    export_path: Path,
    svg: bool,
    png: bool,
    in_place_raster: bool,
):
    cwd = Path(os.getcwd())
    path = (
        export_path.relative_to(cwd)
        if export_path.is_relative_to(cwd)
        else export_path
    )

    if svg:
        logging.info(f"Writing svg: {glyph.glyph_id} -> {path}.svg")
        glyph.export_svg(export_path)

    if png:
        logging.info(f"Writing png: {glyph.glyph_id} -> {path}.png")
        glyph.export_png(export_path, in_place_raster=in_place_raster)


def _extract_containers(fqcn: str) -> list[ExportSpec]:
    """
    Extract all glyphs from the provided FQCN, which may be any of the
    following:

    - BaseGlyph subclass or instance
    - Iterable of the above (subclasses and instances can be intermixed)
    - Callable which returns any of the above
    - Module containing any of the above, with symbol names provided via
      `__all__`
    """

    glyph_specs: list[GlyphSpecType]

    glyph_specs = _import_glyph_specs(fqcn)
    containers: list[ExportSpec] = _normalize_glyph_specs(glyph_specs)

    return containers


def _import_glyph_specs(fqcn: str) -> list[GlyphSpecType]:
    glyph_specs: list[GlyphSpecType]

    module: ModuleType | None = None
    obj: GlyphSpecType | None = None
    import_excep: ImportError | AttributeError | None = None

    try:
        # attempt to import module
        module = importlib.import_module(fqcn)
    except ImportError as e:
        import_excep = e

    if module is None and "." in fqcn:
        # attempt to import object from module
        module_path, obj_name = fqcn.rsplit(".", 1)

        try:
            module = importlib.import_module(module_path)
            obj = getattr(module, obj_name)
        except (ImportError, AttributeError) as e:
            import_excep = e
        else:
            # clear exception as we successfully imported the object
            import_excep = None

    if import_excep is not None:
        logging.error(f"Failed to import object: {fqcn}")
        raise import_excep

    assert module is not None

    glyph_specs: list[Any] = _import_all(module) if obj is None else [obj]
    return cast(list[GlyphSpecType], glyph_specs)


def _normalize_glyph_specs(
    glyph_specs: list[GlyphSpecType],
) -> list[ExportSpec]:
    """
    Take an object and return a list of ExportSpec instances.
    """

    containers: list[ExportSpec] = []

    for glyph_spec in glyph_specs:
        containers_extract = _recurse_glyph_spec(glyph_spec)

        # validate returned objects
        for container in containers_extract:
            assert isinstance(container, ExportSpec)
            containers.append(container)

    return containers


def _recurse_glyph_spec(glyph_spec: GlyphSpecType) -> list[GlyphSpecType]:
    """
    Recurse into glyph spec until we find a glyph class, glyph instance, or
    export spec. A container will be created if not found.
    """

    ret: list[ExportSpec] = []

    if isinstance(glyph_spec, ExportSpec):
        ret.append(glyph_spec)

    elif isinstance(glyph_spec, BaseGlyph):
        ret.append(ExportSpec(glyph_spec, Path()))

    elif isinstance(glyph_spec, Iterable):
        for spec in glyph_spec:
            ret += _recurse_glyph_spec(spec)

    # function, BaseGlyph subclass, or BaseVariantExportFactory subclass
    elif isinstance(glyph_spec, Callable):
        ret += _recurse_glyph_spec(glyph_spec())

    else:
        raise Exception(f"Invalid glyph_spec: {glyph_spec}")

    return ret


def _import_all(module: ModuleType) -> list[Any]:
    all_: list[str] | None = None
    objs: list[Any] = []

    try:
        all_ = module.__all__
    except Exception:
        pass

    assert all_ is not None, f"Module does not have attribute __all__: {module}"

    for attr in all_:
        obj: Any | None = None
        try:
            obj = getattr(module, attr)
        except AttributeError:
            pass

        assert obj is not None, f"Failed to import {attr} from {module}"
        objs.append(obj)

    return objs



from rich.console import Console
from rich.logging import RichHandler
import logging
import rich.traceback
import typer

from ..core.export import export_glyphs

LEVEL = logging.INFO


app = typer.Typer(
    rich_markup_mode="rich",
    no_args_is_help=True,
    add_completion=False,
)

rich.traceback.install(show_locals=True)
logging.basicConfig(
    level=LEVEL,
    format="%(message)s",
    handlers=[
        RichHandler(
            rich_tracebacks=True,
            console=Console(),
            show_level=True,
            show_time=False,
            show_path=False,
            markup=True,
        )
    ],
)


# TODO: take size, dpi
@app.command(no_args_is_help=True)
def export(
    fqcn: str = typer.Argument(help="FQCN of glyph(s) to export"),
    output_path: str = typer.Argument(help="Output path (file or folder)"),
    output_modpath: bool = typer.Option(
        False,
        help="Whether to create subfolders based on the respective glyph's modpath",
    ),
    svg: bool = typer.Option(False, help="Output .svg"),
    png: bool = typer.Option(False, help="Output .png from .svg (Linux only)"),
):
    export_glyphs(
        fqcn, output_path, output_modpath=output_modpath, svg=svg, png=png
    )


def run():
    app()


if __name__ == "__main__":
    run()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Iterable, Self, TypeVar, get_args, cast

from pydantic import BaseModel, ConfigDict
from svgwrite.container import SVG, Group

from .graphics._graphics import (
    GraphicsContainer,
    BaseContainer,
)
from .graphics.properties import Properties

__all__ = [
    "BaseParams",
    "BaseGlyph",
    "EmptyParams",
    "EmptyGlyph",
]


class BaseParams(BaseModel):
    """
    Subclass this class to create parameters for a Glyph.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def desc(self) -> str:
        """
        Short description of params and values.
        """

        def parse_val(val: Any) -> str:
            if isinstance(val, type):
                return val.__name__
            elif isinstance(val, BaseParams):
                return val.desc
            elif isinstance(val, Iterable) and not isinstance(val, str):
                return f"_".join([parse_val(v) for v in val])
            else:
                return str(val).replace("=", "~").replace(".", "_")

        params = []

        for field in type(self).model_fields.keys():
            val: Any = getattr(self, field)
            val_desc = parse_val(val)

            params.append(f"{field}-{val_desc}")

        return "__".join(params)


class BaseGlyph[ParamsT: BaseParams](ABC, GraphicsContainer, BaseContainer):
    """
    Base class for a standalone or reusable glyph, sized in abstract
    (user) units.
    """

    params: ParamsT
    """
    Instance of params with type as specified by typevar.
    """

    DefaultParams: type[ParamsT] | None = None
    """
    Subclass of ParamsT to use as defaults for this particular Glyph,
    if no parameters provided during instantiation.
    """

    _params_cls: type[ParamsT]
    """
    Params class: the type parameter provided as ParamsT, or EmptyParams
    if no type parameter provided.
    """

    def __init__(
        self,
        *,
        glyph_id: str | None = None,
        params: ParamsT | None = None,
        properties: Properties | None = None,
        size: tuple[float, float] | None = None,
        parent: BaseGlyph | None = None,
        insert: tuple[float, float] | None = None,
    ):
        """
        :param parent: Parent glyph, or `None`{l=python} to create top-level glyph
        :param glyph_id: Unique identifier, or `None`{l=python} to generate one
        """

        # if parent not provided, ensure insert/size_inst not provided
        if parent is None:
            assert insert is None

        super().__init__(glyph_id, properties, size)

        # set params
        self.params = params or type(self).get_params_cls()()

        # invoke subclass's init (e.g. set properties based on params)
        self.init()

        # invoke post-init since size_canon may be set in init()
        self._init_post()

        # invoke subclass's drawing logic
        self.draw()

        if parent is not None:
            parent.insert_glyph(self, insert)

    def __init_subclass__(cls: type[BaseGlyph]):
        """
        Populate _params_cls with the class representing the parameters for
        this glyph. If not subscripted with a type arg by the subclass,
        _params_cls is set to EmptyParams.
        """

        # note: pyright is not yet aware that the new style
        #
        # class BaseGlyph[ParamsT: BaseParams](...): ...
        #
        #   is equivalent to
        #
        # ParamsT = TypeVar("ParamsT", bound=BaseParams)
        # class BaseGlyph(Generic[ParamsT], ...): ...
        #
        # since with the new style, Generic is not a base. Nonetheless
        # __orig_bases__ is indeed defined here.
        orig_bases = (
            cls.__orig_bases__  # pyright: ignore[reportGeneralTypeIssues]
        )

        # get_args() returns tuple[Any, ...], but we know the type param
        # should always be ParamsT
        args: tuple[type[ParamsT]] = cast(
            tuple[type[ParamsT]], get_args(orig_bases[0])
        )

        if len(args) > 0:
            assert len(args) == 1

            params_cls: type[ParamsT]

            if isinstance(args[0], TypeVar):
                # have a TypeVar, look up its bound
                type_var = cast(TypeVar, args[0])
                assert type_var.__bound__ is not None
                params_cls = type_var.__bound__
            else:
                # already have a class
                params_cls = args[0]

            cls._params_cls = params_cls
        else:
            # set _params_cls with EmptyParams
            cls._params_cls = EmptyParams

        assert issubclass(cls._params_cls, BaseParams)

    @property
    def glyph_id(self) -> str:
        """
        A meaningful identifier to associate with this glyph. Also used as
        base name (without extension) of file to write when no filename is
        provided.

        If no glyph_id is provided when created, it is derived from the
        class name.
        """
        return self._id_norm

    @classmethod
    def get_params_cls(cls) -> type[ParamsT]:
        """
        Returns the {obj}`BaseParams` subclass with which this class is
        parameterized, accounting for any default values provided by
        the subclass.
        """
        return cls.DefaultParams or cls._params_cls

    def insert_glyph(
        self,
        glyph: BaseGlyph,
        insert: tuple[float, float] | None = None,
    ) -> Self:
        self._nested_glyphs.append(glyph)

        # add group to self, using wrapper svg for placement
        wrapper_insert: SVG = self._drawing.svg(
            **glyph._get_elem_kwargs(suffix="wrapper-insert"),
            insert=insert,
        )

        wrapper_insert.add(glyph._group)
        self._svg.add(wrapper_insert)

        return self

    def init(self):
        ...

    @abstractmethod
    def draw(self):
        ...


class EmptyParams(BaseParams):
    pass


class EmptyGlyph(BaseGlyph[EmptyParams]):
    """
    Glyph to use as an on-the-fly alternative to subclassing
    {obj}`BaseGlyph`. It has an empty {obj}`BaseGlyph.draw`
    implementation; the user can then add graphics objects
    and other glyphs after creation.

    Example:

    ```python
    glyph1 = MyGlyph1()
    glyph2 = MyGlyph2()

    # create an empty glyph with unspecified size
    glyph = EmptyGlyph()

    # insert a glyph
    glyph.insert_glyph(glyph1)

    # insert another glyph in a different position
    glyph.insert_glyph(glyph2, (100, 100))
    ```
    """

    def draw(self):
        pass



from abc import ABC
from typing import Sequence

from ..core import BaseGlyph, BaseParams

__all__ = [
    "ArrayParams",
    "HArrayGlyph",
    "VArrayGlyph",
]


class ArrayParams(BaseParams):
    glyphs: Sequence[BaseGlyph]
    spacing: float = 10.0
    center: bool = False


class BaseArrayGlyph(BaseGlyph[ArrayParams], ABC):
    """
    Glyph encapsulating an array of glyphs with constant spacing between them,
    either horizontal or vertical.

    Horizontal arrays grow to the right, and vertical arrays grow downwards.

    If `center` is `True`, glyphs are center aligned.
    """

    _vertical: bool = False

    @classmethod
    def new(
        cls,
        glyphs: Sequence[BaseGlyph],
        glyph_id: str | None = None,
        spacing: float = 0.0,
        center: bool = False,
    ):
        params = cls.get_params_cls()(
            glyphs=glyphs, spacing=spacing, center=center
        )
        return cls(params=params, glyph_id=glyph_id)

    def init(self):
        self.size_canon = self._get_size()

    def draw(self):
        # set initial insert point
        insert: tuple[float, float] = (0.0, 0.0)

        for glyph in self.params.glyphs:
            # adjust insert point for this glyph
            insert = self._align_insert(insert, glyph)

            # insert glyph
            self.insert_glyph(glyph, insert)

            # advance insert point
            insert = self._advance_insert(insert, glyph)

    @property
    def _x_center(self) -> float:
        assert self.size_canon is not None
        return self.size_canon[0] / 2

    @property
    def _y_center(self) -> float:
        assert self.size_canon is not None
        return self.size_canon[1] / 2

    def _get_size(self) -> tuple[float, float]:
        width: float
        height: float

        if self._vertical:
            # width: widest glyph
            # height: sum of heights and spacings
            width = max(g.width for g in self.params.glyphs)
            height = self._get_size_dim(1)
        else:
            # width: sum of widths and spacings
            # height: tallest glyph
            width = self._get_size_dim(0)
            height = max(g.height for g in self.params.glyphs)

        return (width, height)

    def _get_size_dim(self, dim: int) -> float:
        """
        Get size along provided dimension. Assumes the array is oriented along
        the provided dimension (e.g. 0 for horizontal).
        """

        length: float = 0.0

        for glyph in self.params.glyphs:
            length += glyph.size[dim]

        # add spacing for all except one
        length += self.params.spacing * (len(self.params.glyphs) - 1)

        return length

    def _align_insert(
        self, insert: tuple[float, float], glyph: BaseGlyph
    ) -> tuple[float, float]:
        """
        Center this glyph to the starting point in the direction of the
        array.
        """

        x: float
        y: float

        if self.params.center:
            # compute one axis of this insertion relative to start, centering
            # the glyph along the appropriate axis
            if self._vertical:
                # align x relative to start, preserve y
                x = self._x_center - (glyph.size[0] / 2)
                y = insert[1]
            else:
                # align y relative to start, preserve x
                x = insert[0]
                y = self._y_center - (glyph.size[1] / 2)
        else:
            # no need to center, just align along the needed dimension
            if self._vertical:
                x = 0.0
                y = insert[1]
            else:
                x = insert[0]
                y = 0.0

        return (x, y)

    def _advance_insert(
        self, insert: tuple[float, float], glyph: BaseGlyph
    ) -> tuple[float, float]:
        if self._vertical:
            # preserve x, advance y
            return (insert[0], insert[1] + glyph.size[1] + self.params.spacing)
        else:
            # preserve y, advance x
            return (insert[0] + glyph.size[0] + self.params.spacing, insert[1])


class HArrayGlyph(BaseArrayGlyph):
    _vertical = False


class VArrayGlyph(BaseArrayGlyph):
    _vertical = True



import string
import sys

from pydantic import Field

from ....core.glyph import BaseGlyph, BaseParams
from ....core.graphics import Properties, PropertyValue
from ...utils import extend_line


LETTERS_STR = string.ascii_uppercase
"""
Assume there are classes in this module with single-letter names A-Z.
"""

__all__ = [
    "LetterParams",
    "BaseLetterGlyph",
    "LetterComboParams",
    "BaseLetterComboGlyph",
    "ZERO",
    "UNIT",
    "HALF",
    "LETTER_CLS_LIST",
    *LETTERS_STR,
]

ZERO = 0.0
UNIT = 100.0
HALF = UNIT / 2
QUART = HALF / 2


class LetterParams(BaseParams):
    color: str = "black"
    stroke_pct: float = 5.0


class BaseLetterGlyph(BaseGlyph[LetterParams]):
    _stroke_width: float

    class DefaultProperties(Properties):
        fill: PropertyValue = "none"
        stroke: PropertyValue = "black"
        stroke_linejoin: PropertyValue = "bevel"

    size_canon = (UNIT, UNIT)

    def init(self):
        self._stroke_width = (self.params.stroke_pct / 100) * UNIT
        self.properties.stroke_width = str(round(self._stroke_width))
        self.properties.stroke = self.params.color

    @property
    def _stroke_half(self) -> float:
        return self._stroke_width / 2

    @property
    def _stroke_start(self) -> float:
        return self._stroke_half

    @property
    def _stroke_end(self) -> float:
        return UNIT - self._stroke_half


class LetterComboParams(BaseParams):
    """
    Contains parameters to propagate to letters.
    """

    letter_params: LetterParams = Field(default_factory=LetterParams)


class BaseLetterComboGlyph[ParamsT: LetterComboParams](BaseGlyph[ParamsT]):
    """
    Glyph which encapsulates a combination of overlayed letter glyphs.
    """

    size_canon = (UNIT, UNIT)

    def draw_letter[
        LetterT: BaseLetterGlyph
    ](self, letter_cls: type[LetterT]) -> LetterT:
        return letter_cls(parent=self, params=self.params.letter_params)

    def draw_combo[
        ComboT: BaseLetterComboGlyph
    ](self, combo_cls: type[ComboT]) -> ComboT:
        return combo_cls(parent=self, params=self.params)


class A(BaseLetterGlyph):
    def draw(self):
        # top point
        top = (HALF, ZERO)

        self.draw_polyline(
            [
                extend_line((self._stroke_start, UNIT), top),
                top,
                extend_line((self._stroke_end, UNIT), top),
            ]
        )

        self.draw_polyline(
            [
                (QUART, HALF),
                (QUART * 3, HALF),
            ]
        )


class B(BaseLetterGlyph):
    def draw(self):
        ...


class C(BaseLetterGlyph):
    def draw(self):
        ...


class D(BaseLetterGlyph):
    def draw(self):
        ...


class E(BaseLetterGlyph):
    def draw(self):
        ...


class F(BaseLetterGlyph):
    def draw(self):
        self.draw_polyline(
            [(self._stroke_half, ZERO), (self._stroke_half, UNIT)]
        )
        self.draw_polyline(
            [(ZERO, self._stroke_half), (UNIT, self._stroke_half)]
        )
        self.draw_polyline([(ZERO, HALF), (UNIT, HALF)])


class G(BaseLetterGlyph):
    def draw(self):
        ...


class H(BaseLetterGlyph):
    def draw(self):
        ...


class I(BaseLetterGlyph):
    def draw(self):
        self.draw_polyline([(HALF, ZERO), (HALF, UNIT)])


class J(BaseLetterGlyph):
    def draw(self):
        ...


class K(BaseLetterGlyph):
    def draw(self):
        ...


class L(BaseLetterGlyph):
    def draw(self):
        ...


class M(BaseLetterGlyph):
    def draw(self):
        self.draw_polyline(
            [
                (self._stroke_half, UNIT),
                (self._stroke_half, ZERO),
                (HALF, UNIT),  # TODO: determine ideal y-coordinate
                (self._stroke_end, ZERO),
                (self._stroke_end, UNIT),
            ]
        )


class N(BaseLetterGlyph):
    def draw(self):
        ...


class O(BaseLetterGlyph):
    def draw(self):
        ...


class P(BaseLetterGlyph):
    def draw(self):
        ...


class Q(BaseLetterGlyph):
    def draw(self):
        ...


class R(BaseLetterGlyph):
    def draw(self):
        ...


class S(BaseLetterGlyph):
    def draw(self):
        ...


class T(BaseLetterGlyph):
    def draw(self):
        self.draw_polyline(
            [
                (ZERO, self._stroke_start),
                (UNIT, self._stroke_start),
                (HALF, self._stroke_start),
                (HALF, UNIT),
            ]
        )


class U(BaseLetterGlyph):
    def draw(self):
        ...


class V(BaseLetterGlyph):
    def draw(self):
        ...


class W(BaseLetterGlyph):
    def draw(self):
        ...


class X(BaseLetterGlyph):
    def draw(self):
        ...


class Y(BaseLetterGlyph):
    def draw(self):
        ...


class Z(BaseLetterGlyph):
    def draw(self):
        ...


LETTER_CLS_LIST: list[type[BaseLetterGlyph]] = [
    getattr(sys.modules[__name__], l) for l in LETTERS_STR
]
"""
List of letter classes in order.
"""



from __future__ import annotations

import logging
from typing import Any, cast

from svgwrite.drawing import Drawing
from svgwrite.container import SVG, Group

from .properties import Properties


class BaseContainer:
    """
    Container for drawing and manipulation of low-level graphics objects.

    There are 2 SVG trees maintained:
    1. Standalone glyph
        drawing (root <svg>)
            wrapper svg (if size provided)
                canonical svg
    2. Placement in parent glyph
        svg-parent
            svg-wrapper-placement (insert only)
                group (w/transforms, e.g. rotation)
                    svg-wrapper-scaling (width/height/viewbox only)
                        svg-canonical
    """

    DefaultProperties: type[Properties] = Properties
    """
    Subclass of Properties with properties set as class attributes.
    
    May be subclassed in-place or assigned to an existing subclass.
    For example, subclassing in-place:

    ```python
    class MyGlyph(BaseGlyph):

        class DefaultProperties(Properties):
            color: PropertyValue = "black"
            stroke_width: PropertyValue = "10"
        
        ...
    ```
    """

    properties: Properties
    """
    Properties automatically propagated to graphics objects in draw_*() APIs.
    """

    size_canon: tuple[float, float] | None = None
    """
    Canonical size in user units, as provided by concrete class either as class
    attribute or upon creation in {obj}`BaseGlyph.init`.

    Used to set a consistent size for drawing, invariant of the size 
    passed upon instantiation.

    May be `None`, in which case {obj}`BaseContainer.size`, 
    {obj}`BaseContainer.width`, and {obj}`BaseContainer.height`
    will return `None`.

    Note that some {obj}`BaseGlyph` subclasses may require that 
    this field is not `None`.
    """

    viewbox_canon: tuple[tuple[float, float], tuple[float, float]] | None = None

    _id: str | None
    """
    Unique id for this container.
    """

    _size: tuple[float, float] | None = None
    """
    Instantiated size, set upon glyph creation. If `None`, not rescaled;
    uses canonical size.
    """

    _drawing: Drawing
    """
    Drawing object, unique to each Glyph.
    """

    _svg: SVG
    """
    Canonical SVG container for all elements generated by this glyph.
    """

    _group: Group
    """
    Group to hold wrapper SVG container when added to a parent glyph.
    Transformations are performed on this group rather than the <svg> itself
    as <svg> does not support transformations in SVG 1.1.
    """

    _nested_glyphs: list[BaseContainer] = []
    """
    List of glyphs nested under this one, mostly for debugging.
    """

    def __init__(
        self,
        id_: str | None,
        properties: Properties | None,
        size: tuple[float, float] | None,
    ):
        self._id = id_
        self._size = size
        self._drawing = Drawing()
        self._group = self._drawing.g(**self._get_elem_kwargs(suffix="group"))
        self._nested_glyphs = []
        self.properties = properties or self.DefaultProperties()

    @property
    def has_size(self) -> bool:
        try:
            _ = self.size
        except AssertionError:
            return False

        return True

    # TODO: document: raises exception + add custom exception
    @property
    def size(self) -> tuple[float, float]:
        s = self._size or self.size_canon
        assert s is not None
        return s

    @property
    def width(self) -> float:
        return self.size[0]

    @property
    def height(self) -> float:
        return self.size[1]

    @property
    def _id_norm(self) -> str:
        return self._id or type(self).__name__

    def _get_elem_kwargs(self, suffix: str | None = None) -> dict[str, str]:
        kwargs: dict[str, str] = {}
        suffix_ = "" if suffix is None else f"-{suffix}"

        if self._id:
            kwargs["id_"] = f"{self._id}{suffix_}"

        kwargs["class_"] = f"{type(self).__name__}{suffix_}"

        return kwargs

    def _init_post(self):
        # create canonical svg
        self._svg = cast(
            SVG,
            self._drawing.svg(
                **self._get_elem_kwargs(),
                size=self.size_canon,
            ),
        )

        # set viewbox, if configured
        if self.viewbox_canon is not None:
            x, y = self.viewbox_canon[0]
            w, h = self.viewbox_canon[1]
            self._svg.viewbox(x, y, w, h)
            self._svg.fit()

        # add to drawing for standalone glyph
        if self._size is None:
            # no scaling needed, add svg directly
            self._drawing.add(self._svg)
            self._group.add(self._svg)
        else:
            # create wrapper svg and rescale
            svg_wrapper: SVG = self._create_wrapper_scale()
            self._drawing.add(svg_wrapper)
            self._group.add(svg_wrapper)

        # set top-level dimensions explicitly as larger SVGs
        # are unexpectedly truncated when sized to 100% (default)
        if self.has_size:
            self._drawing["width"] = str(self.size[0])
            self._drawing["height"] = str(self.size[1])

    def _create_wrapper_scale(self) -> SVG:
        """
        Create SVG wrapper for canonical SVG object to handle
        placement/scaling.
        """

        size_kwargs = {} if self._size is None else {"size": self._size}
        wrapper_scale = cast(
            SVG,
            self._drawing.svg(
                **self._get_elem_kwargs(suffix="wrapper-scale"),
                **size_kwargs,
            ),
        )
        self._rescale_svg(wrapper_scale, self._size)
        wrapper_scale.add(self._svg)

        return wrapper_scale

    def _rescale_svg(
        self,
        svg: SVG,
        size: tuple[float, float] | tuple[str, str] | None,
        set_size: bool = False,
    ):
        if size is not None:
            if set_size:
                svg["width"] = str(size[0])
                svg["height"] = str(size[1])

            if self.size_canon is None:
                logging.warning(
                    f"Rescaling glyph {self} which has no canonical size; viewbox will not be set, so results may be unexpected"
                )
            else:
                svg.viewbox(
                    0, 0, round(self.size_canon[0]), round(self.size_canon[1])
                )
                svg.fit()



from __future__ import annotations

from collections.abc import Iterable
from typing import Self

from svgwrite.container import Group

from svgwrite.shapes import (
    Line,
    Rect,
    Circle,
    Ellipse,
    Polyline,
    Polygon,
)

from ._container import BaseContainer
from .properties import BaseProperties, PaintingProperties


class DrawContainer(BaseContainer):
    def draw_line(
        self,
        start: tuple[float, float],
        end: tuple[float, float],
        **extra: str,
    ) -> Line:
        elem: Line = self._drawing.line(
            start=start,
            end=end,
            **self._get_extra_painting(extra),
        )

        self._svg.add(elem)

        return elem

    def draw_polyline(
        self,
        points: Iterable[tuple[float, float]],
        **extra: str,
    ) -> Polyline:
        elem: Polyline = self._drawing.polyline(
            points=[p for p in points],
            **self._get_extra_painting(extra),
        )

        self._svg.add(elem)

        return elem

    def draw_polygon(
        self,
        points: Iterable[tuple[float, float]],
        **extra: str,
    ) -> Polygon:
        elem: Polygon = self._drawing.polyline(
            points=[p for p in points],
            **self._get_extra_painting(extra),
        )

        self._svg.add(elem)

        return elem

    def draw_rect(
        self,
        insert: tuple[float, float],
        size: tuple[float, float],
        radius_x: float | None = None,
        radius_y: float | None = None,
        **extra: str,
    ) -> Rect:
        elem: Rect = self._drawing.rect(
            insert=insert,
            size=size,
            rx=radius_x,
            ry=radius_y,
            **self._get_extra_painting(extra),
        )

        self._svg.add(elem)

        return elem

    def draw_circle(
        self,
        center: tuple[float, float],
        radius: float,
        **extra: str,
    ) -> Circle:
        elem: Circle = self._drawing.circle(
            center=center, r=radius, **self._get_extra_painting(extra)
        )

        self._svg.add(elem)

        return elem

    def draw_ellipse(
        self,
        center: tuple[float, float],
        radius: tuple[float, float],
        **extra: str,
    ) -> Ellipse:
        elem: Ellipse = self._drawing.ellipse(
            center=center, r=radius, **self._get_extra_painting(extra)
        )

        self._svg.add(elem)

        return elem

    def draw_group(self, **extra: str) -> Group:
        elem: Group = self._drawing.g(**extra)

        self._svg.add(elem)

        return elem

    def _get_extra(
        self, extra: dict[str, str], *extra_cls: type[BaseProperties]
    ) -> dict[str, str]:
        # get extras defined on each class
        valid_extras = []
        for cls in extra_cls:
            valid_extras += list(cls.model_fields.keys())

        for attr in extra:
            assert attr in valid_extras, f"Invalid extra {attr}"

        ret = extra.copy()

        # check for extra attributes defined on class
        for attr in valid_extras:
            # explicitly provided extra takes precedence
            if attr in extra:
                continue

            if (value := getattr(self.properties, attr)) is not None:
                ret[attr] = value

        return ret

    def _get_extra_painting(self, extra: dict[str, str]) -> dict[str, str]:
        return self._get_extra(extra, PaintingProperties)


# TODO: other methods of https://svgwrite.readthedocs.io/en/latest/classes/mixins.html#svgwrite.mixins.Transform
class TransformContainer(BaseContainer):
    def rotate(
        self, angle: float, center: tuple[float, float] | None = None
    ) -> Self:
        # set center if none was provided and we have a size
        if center is None and self.size is not None:
            center = (self.size[0] / 2, self.size[1] / 2)

        self._group.rotate(angle, center=center)

        return self

    def scale(self, x: float, y: float | None = None) -> Self:
        if y is None:
            y = x

        self._group.scale(x, y)

        return self



"""
Functionality to create glyphs with user-defined parameter variations.
For visualization, all variants can be represented as a matrix and a
set of arrays.
"""

from pathlib import Path
from typing import Generator

from ..core.export import ExportSpec

from glyphsynth import (
    BaseGlyph,
    BaseParams,
    HArrayGlyph,
    VArrayGlyph,
    PaddingGlyph,
)

__all__ = [
    "BaseVariantFactory",
    "BaseVariantExportFactory",
]


class BaseVariantFactory[GlyphT: BaseGlyph]:
    """
    Encapsulates a BaseGlyph subclass and variants for parameters and
    properties.
    """

    glyph_cls: type[GlyphT]
    """
    BaseGlyph subclass to instantiate.
    """

    def get_variants(self) -> Generator[GlyphT, None, None]:
        """
        Yield all variants.
        """
        for params in self.get_params_variants():
            glyph_id = _derive_glyph_id(params)
            yield self.glyph_cls(glyph_id=glyph_id, params=params)

    def get_params_variants(self) -> Generator[BaseParams, None, None]:
        """
        Override to yield parameter variants to export.
        """
        yield self.glyph_cls.get_params_cls()()


class BaseVariantExportFactory[GlyphT: BaseGlyph](BaseVariantFactory[GlyphT]):
    """
    Encapsulates a variant factory which creates and exports arrays of
    glyphs with combinations of glyph parameters and properties.

    Exports a hierarchy of glyph variants:

    - all/[glyph_id].[svg/png]
    - matrix/
        - All glyphs combined in matrix
    - harrays/
        - Horizontal arrays
    - varrays/
        - Vertical arrays
    """

    MATRIX_WIDTH: int = 1
    """
    Width of the resulting matrix glyph. If kept as the default of 1, creates a
    vertical array.

    If set, should equal the number of elements in outermost dimension 
    of the glyph iterable output.

    If a dynamic value is required, override the property `matrix_width`.
    """

    SPACING: int = 0
    """
    Spacing to use between variants.
    """

    def __iter__(self) -> Generator[ExportSpec, None, None]:
        """
        Yield export specs for each variant given by
        concrete VariantFactory.
        """

        def wrap_padding(glyph: BaseGlyph):
            return PaddingGlyph.new(glyph, padding=self.SPACING)

        # get list of all glyphs
        all_glyphs: list[GlyphT] = list(self.get_variants())

        # length of slices to create as vertical arrays
        v_len = len(all_glyphs) // self.matrix_width

        # length of slices to create as horizontal arrays
        h_len = self.matrix_width

        # list of horizontal arrays
        harrays: list[list[GlyphT]] = [[] for _ in range(v_len)]
        harray_glyphs: list[HArrayGlyph] = []

        # list of vertical arrays
        varrays: list[list[GlyphT]] = [[] for _ in range(v_len)]
        varray_glyphs: list[VArrayGlyph] = []

        # relative paths for exporting
        variants_path = Path("variants")
        all_path = variants_path / "all"
        harrays_path = variants_path / "harrays"
        varrays_path = variants_path / "varrays"
        matrix_path = variants_path / "matrix"

        # create horizontal/vertical arrays
        for i in range(v_len):
            harrays[i] = [all_glyphs[j * v_len + i] for j in range(h_len)]

        for i in range(h_len):
            varrays[i] = all_glyphs[i * v_len : (i + 1) * v_len]

        # create array glyphs from raw arrays
        for i, harray in enumerate(harrays):
            if len(harray) == 0:
                continue
            harray_glyphs.append(
                HArrayGlyph.new(
                    harray, glyph_id=f"row_{i}", spacing=self.SPACING
                )
            )

        for i, varray in enumerate(varrays):
            if len(varray) == 0:
                continue

            varray_glyphs.append(
                VArrayGlyph.new(
                    varray, glyph_id=f"column_{i}", spacing=self.SPACING
                )
            )

        # export top-level glyphs
        for glyph in all_glyphs:
            yield ExportSpec(
                wrap_padding(glyph), all_path, module=type(self).__module__
            )

        # export horizontal arrays
        for i, harray_glyph in enumerate(harray_glyphs):
            yield ExportSpec(
                wrap_padding(harray_glyph),
                harrays_path,
                module=type(self).__module__,
            )

        # export vertical arrays
        for i, varray_glyph in enumerate(varray_glyphs):
            yield ExportSpec(
                wrap_padding(varray_glyph),
                varrays_path,
                module=type(self).__module__,
            )

        # export matrix glyph as vertical array of horizontal arrays
        yield ExportSpec(
            wrap_padding(
                VArrayGlyph.new(
                    harray_glyphs, glyph_id="matrix", spacing=self.SPACING
                )
            ),
            matrix_path,
            module=type(self).__module__,
        )

    @property
    def matrix_width(self) -> int:
        """
        Accessor for the attribute `MATRIX_WIDTH`, but can be overridden if
        a dynamic value is needed.
        """
        return self.MATRIX_WIDTH


def _derive_glyph_id(params: BaseParams) -> str:
    """
    Derive a glyph_id from params.
    """
    return params.desc



from __future__ import annotations

from ._draw import DrawContainer, TransformContainer
from ._export import ExportContainer
from ._container import BaseContainer


class GraphicsContainer(
    DrawContainer, TransformContainer, ExportContainer, BaseContainer
):
    pass



from __future__ import annotations

from pathlib import Path
from shutil import which
from typing import Literal
from xml.etree import ElementTree
import copy
import logging
import subprocess
import os
import shutil
import sys
import tempfile
import xml.dom.minidom as minidom

from svgwrite.drawing import Drawing

from . import RASTER_SUPPORT
from ._container import BaseContainer


class ExportContainer(BaseContainer):
    def export(
        self, path: Path, out_format: Literal["svg", "png"] | None = None
    ):
        match self._get_format(path, out_format):
            case "svg":
                self.export_svg(path)
            case "png":
                # TODO: pass through png-specific params?
                self.export_png(path)
            case _:
                raise Exception

    def export_svg(self, path: Path):
        """
        :param path: Path to destination file or folder
        """

        path_norm: Path = self._normalize_path(path, "svg")

        with path_norm.open("w") as fh:
            fh.write(self._get_svg())

    def export_png(
        self,
        path: Path,
        size: tuple[str, str] | None = None,
        dpi: tuple[int, int] = (96, 96),
        scale: float | int = 1,
        in_place_raster: bool = False,
    ):
        """
        :param path: Path to destination file or folder
        :param size: Size of image with concrete units (px/in/...), e.g. `("1in", "1in")`{l=python}, or `None`{l=python} to use provided scale factor
        :param dpi: Pixels per inch
        :param scale: Factor by which to scale user units to concrete pixels, only if `size is None`{l=python}
        """

        path_norm: Path = self._normalize_path(path, "png")
        size_raster: tuple[str, str] = size or self._get_size_raster(
            float(scale)
        )

        self._rasterize(path_norm, size_raster, dpi, in_place_raster)

    def get_svg(self) -> str:
        return self._get_svg()

    def _get_svg(self, drawing: Drawing | None = None) -> str:
        """
        Get a string containing the full XML content.
        """

        # if no drawing provided, default to drawing for this glyph
        drawing_: Drawing = drawing or self._drawing

        # get element as string
        xml_str: str = ElementTree.tostring(
            drawing_.get_xml(), encoding="utf-8", xml_declaration=True
        ).decode("utf-8")

        xml_tree = minidom.parseString(xml_str)
        return xml_tree.toprettyxml(indent="  ")

    def _rasterize(
        self,
        path_png: Path,
        size_raster: tuple[str, str],
        dpi: tuple[int, int],
        in_place_raster: bool,
    ):
        # ensure rsvg-convert is supported and available
        if not RASTER_SUPPORT:
            sys.exit(
                "Conversion to .png only supported on Linux due to availability of rsvg-convert"
            )

        if (path_rsvg_convert := which("rsvg-convert")) is None:
            sys.exit("Could not find path to rsvg-convert")

        logging.debug(f"Found path to rsvg-convert: {path_rsvg_convert}")

        # create temp svg file scaled appropriately
        path_svg: Path

        path_svg_dir = (
            path_png.parent if in_place_raster else Path(tempfile.mkdtemp())
        )

        path_svg = path_svg_dir / f"{path_png.name}.temp.svg"
        self._create_svg_temp(path_svg, size_raster)

        logging.debug(
            f"Rasterizing: {path_svg} -> {path_png}, size_raster={size_raster}, dpi={dpi}"
        )

        # TODO: pass background color in API?
        args = [
            path_rsvg_convert,
            "--keep-aspect-ratio",
            "--background-color",
            "#ffffff",
            "--dpi-x",
            f"{dpi[0]}",
            "--dpi-y",
            f"{dpi[1]}",
            "-o",
            str(path_png),
            str(path_svg),
        ]

        logging.debug(f"Running: {' '.join(args)}")

        subprocess.check_call(args)

        # clean up temp dir
        if not in_place_raster:
            shutil.rmtree(path_svg.parent)

    def _create_svg_temp(self, path_svg: Path, size_raster: tuple[str, str]):
        """
        Create temp .svg for rasterizing.
        """

        # create temp drawing (top-level <svg>) and set size in order to
        # set output size
        # - required even if size provided to rsvg-convert
        drawing_tmp: Drawing = copy.copy(self._drawing)

        self._rescale_svg(drawing_tmp, size_raster, set_size=True)

        svg_str = self._get_svg(drawing_tmp)

        with path_svg.open("w") as fh:
            fh.write(svg_str)

    def _get_size_raster(self, scale: float):
        assert (
            self.size is not None
        ), f"No size provided for rasterizing and glyph {self} has no size"

        x = int(self.size[0] * scale)
        y = int(self.size[1] * scale)

        return (f"{x}px", f"{y}px")

    def _normalize_path(
        self, path: Path, out_format: Literal["svg", "png"] | None
    ) -> Path:
        """
        Take path (folder or file) and return complete filename.
        """

        out_format_norm = self._get_format(path, out_format)
        is_dir: bool

        if path.exists():
            # existing path, determine if directory
            is_dir = path.is_dir()
        else:
            # not existing path, use heuristics to determine if path is
            # a directory: if it doesn't have an extension, it should be
            # a directory

            is_dir = len(path.suffix) == 0
            path_dir = path if is_dir else path.parent

            # create dirs if needed
            path_dir.mkdir(parents=True, exist_ok=True)

        return path / f"{self._id_norm}.{out_format_norm}" if is_dir else path

    def _get_format(self, path: Path, format: str | None) -> str:
        # check provided format
        if format is not None:
            return format

        # check path
        if len(path.suffix):
            # omit "."
            return path.suffix[1:]

        # default to svg
        return "svg"



from __future__ import annotations

from pydantic import BaseModel

__all__ = [
    "Properties",
    "PropertyValue",
]

type PropertyValue = str | None


class BaseProperties(BaseModel):
    """
    Encapsulates graphics properties, as defined here:
    <https://www.w3.org/TR/SVG11/intro.html#TermProperty>

    And listed here: <https://www.w3.org/TR/SVG11/propidx.html>
    """


class PaintingProperties(BaseProperties):
    color: PropertyValue = None
    color_interpolation: PropertyValue = None
    color_interpolation_filters: PropertyValue = None
    color_profile: PropertyValue = None
    color_rendering: PropertyValue = None
    fill: PropertyValue = None
    fill_opacity: PropertyValue = None
    fill_rule: PropertyValue = None
    image_rendering: PropertyValue = None
    marker: PropertyValue = None
    marker_end: PropertyValue = None
    marker_mid: PropertyValue = None
    marker_start: PropertyValue = None
    shape_rendering: PropertyValue = None
    stroke: PropertyValue = None
    stroke_dasharray: PropertyValue = None
    stroke_dashoffset: PropertyValue = None
    stroke_linecap: PropertyValue = None
    stroke_linejoin: PropertyValue = None
    stroke_miterlimit: PropertyValue = None
    stroke_opacity: PropertyValue = None
    stroke_width: PropertyValue = None
    text_rendering: PropertyValue = None


class FontProperties(BaseProperties):
    font: PropertyValue = None
    font_family: PropertyValue = None
    font_size: PropertyValue = None
    font_size_adjust: PropertyValue = None
    font_stretch: PropertyValue = None
    font_style: PropertyValue = None
    font_variant: PropertyValue = None
    font_weight: PropertyValue = None


class Properties(PaintingProperties, FontProperties):
    """
    Class to represent all styling properties:
    <https://www.w3.org/TR/SVG11/styling.html#SVGStylingProperties>
    """

    def __init_subclass__(cls):
        super().__init_subclass__()

        valid_properties = Properties.model_fields.keys()

        # ensure user didn't add any invalid properties
        for field in cls.model_fields.keys():
            assert field in valid_properties, f"{field} is not a valid property"



from typing import Literal

from ..core import BaseGlyph, BaseParams


__all__ = [
    "PaddingSpec",
    "PaddingParams",
    "PaddingGlyph",
    "extend_line",
]

type SideSpec = Literal["top", "bottom", "left", "right"]

type PaddingSpec = dict[SideSpec, float]

# TODO: get from SideSpec to avoid redundancy
SIDES: list[SideSpec] = ["top", "bottom", "left", "right"]


class PaddingParams(BaseParams):
    glyph: BaseGlyph
    padding: float | PaddingSpec


class PaddingGlyph(BaseGlyph[PaddingParams]):
    _padding: PaddingSpec

    @classmethod
    def new(cls, glyph: BaseGlyph, padding: float | PaddingSpec | None = None):
        padding_: float | PaddingSpec

        # default padding is 10% of the minimum of width/height
        padding_ = (
            min(glyph.width, glyph.height) / 10 if padding is None else padding
        )

        # create params
        params = cls.get_params_cls()(glyph=glyph, padding=padding_)

        # generate glyph id
        glyph_id_ = f"{glyph.glyph_id}-pad"

        return cls(glyph_id=glyph_id_, params=params)

    def init(self):
        self._padding = self._get_padding()
        self.size_canon = self._get_size()

    def draw(self):
        self.insert_glyph(
            self.params.glyph, (self._padding["left"], self._padding["top"])
        )

    def _get_padding(self) -> PaddingSpec:
        padding: PaddingSpec

        if isinstance(self.params.padding, dict):
            padding = self.params.padding.copy()
        else:
            padding = {side: float(self.params.padding) for side in SIDES}

        for side in padding:
            assert side in SIDES, f"Invalid side: {side}"

        for side in SIDES:
            if side not in padding:
                padding[side] = 0.0

        return padding

    def _get_size(self) -> tuple[float, float]:
        width = float(
            self.params.glyph.size[0]
            + self._padding["left"]
            + self._padding["right"]
        )
        height = float(
            self.params.glyph.size[1]
            + self._padding["top"]
            + self._padding["bottom"]
        )

        return (width, height)


# TODO: caption glyph
# - optional custom caption; default based on class name and params


def extend_line(
    point: tuple[float, float], ref: tuple[float, float], scale: float = 1.0
) -> tuple[float, float]:
    """
    Convenience function to return a point along a line collinear with
    the provided point and a reference point.

    The distance between `point` and the returned point is the distance between
    `point` and `ref` scaled by the provided `scale`.
    """

    offset: tuple[float, float] = (point[0] - ref[0], point[1] - ref[1])
    offset_scale: tuple[float, float] = (offset[0] * scale, offset[1] * scale)

    return (
        point[0] + offset_scale[0],
        point[1] + offset_scale[1],
    )



import os

from pyrollup import rollup

from . import properties
from .properties import *

__all__ = ["RASTER_SUPPORT"] + rollup(properties)

RASTER_SUPPORT: bool = os.name == "posix"
"""
Whether rasterization is supported; Linux only.
"""

# TODO: add types.py?
# - Coordinate
# - Box/Area
